//modified by Jing Yang after 7/6/2001
#include <math.h>
#include "CDimReducer.H"
#include "XmdvTool.H"
#include "OnlineDimClustering.H"
#include <string.h>
#ifndef _WIN32
#include <unistd.h>
#endif

CDimReducer::CDimReducer()
{
    data_array = NULL;
    entries_array = NULL;
	dim_tree = NULL;
	representative_dim_list = NULL;

    data_strategy = USE_CLUSTERS;
    tree_strategy = REVERSE_DATACLUSTERING;
    status = NORMAL;

	extent_scaling_fac=.3;
	density = 0.5;
}


CDimReducer::~CDimReducer()
{
    if ( dim_tree != NULL )
        delete dim_tree;
    dtree = NULL;
    if ( representative_dim_list != NULL )
        delete representative_dim_list;

    if ( data_array != NULL )
    {
        for ( int i = 0; i < data_num; i++ )
        {
            delete [] data_array[i];
        }
        delete [] data_array;
    }

    if ( entries_array != NULL )
    {
        for ( int i = 0; i < data_num; i++ )
        {
            delete [] entries_array[i];
        }
        delete [] entries_array ;
    }
}


bool CDimReducer::GetData()
{
    if ( data_array != NULL )
    {
        for ( int i = 0; i < data_num; i++ )
        {
            delete [] data_array[i];
        }
        delete [] data_array;
    }
    if ( entries_array != NULL )
    {
        for ( int i = 0; i < data_num; i++ )
        {
            delete [] entries_array[i];
        }
        delete [] entries_array ;
    }

    switch (data_strategy) 
    {
        case USE_CLUSTERS:
        default: 
            return GetDataFromClusters();
    }
}


void CDimReducer::CollectClustersOnContour(double contour_radius, Cluster node, CLinkList* data_linkList, CLinkList* entries_linkList)
{
    int i;
    if ( node.num_children == 0 || node.sqrt_radius < contour_radius )
    {
	    double* data = new double [dataset->dims];
        int *entries = new int[1];

        ARR<double> mapped_data;
        mapped_data.SetLen(dataset->dims);

        for ( i=0; i < dataset->dims; i++)
        {   
            dataset->MapData(node.mean, mapped_data);
			data[i] = mapped_data[i];
        }

        entries[0] = node.entries;
        data_linkList->AddNodeAtEnd(data);
        entries_linkList->AddNodeAtEnd(entries);
        return;
    }
    else
    {
        for ( i = 0; i < node.num_children; i++ )
        {
            CollectClustersOnContour(contour_radius, btree->nodes[node.child[i]], data_linkList, entries_linkList);
        }
        return;
    }
}


bool CDimReducer::GetDataFromClusters()
{
    double contour_radius = 0.00;
    CLinkList* data_linkList = new CLinkList();
    CLinkList* entries_linkList = new CLinkList();
 /*   if ( hier_info->database )
    {
 	    prepare( 0, 1, contour_radius, NO_P_ENTRIES-1);

	    ENTRY *entry;
	    for (init(); (entry=get_one())!=NULL; )
	    {
		    double* data = new double [dataset->all_dims];
            for (int i=0; i<dataset->all_dims; i++)
            {
			    data[i] = entry->data->D[i];
            }
            data_linkList->AddNodeAtEnd(data);
	    }
    }
    else
  */  {
        if ( ! dataset )
            return false;
        
        hierInitClusterTree();
        if ( !btree )
            return false;
        CollectClustersOnContour(contour_radius, btree->nodes[0], data_linkList, entries_linkList);
    }

    data_array = (double**) data_linkList->TransToArray();
    entries_array = (int**) entries_linkList->TransToArray();
    data_num = data_linkList->GetLength();
    delete data_linkList;
    delete entries_linkList;
    return true;
}


void CDimReducer::RefineData()
{
    if ( data_num <= 0 )
        return;

    double *min, *max, *diff;

    min = new double [dataset->dims];
    max = new double [dataset->dims];
    diff = new double [dataset->dims];

    int i, j;
    for ( j = 0; j < dataset->dims; j++ )
    {
        min[j] = max[j] = data_array[0][j];
        for ( i = 1; i < data_num; i++ )
        {
            if ( min[j] > data_array[i][j] )
                min[j] = data_array[i][j];
            else if ( max[j] < data_array[i][j] )
                max[j] = data_array[i][j];
        }

        if ( fabs(max[j]-min[j]) < 1e-5 )
            min[j] = max[j] - 0.05;
        diff[j] = 1/(max[j]-min[j]);

        for ( i = 0; i < data_num; i++ )
        {
            data_array[i][j] = (data_array[i][j]-min[j])*diff[j];
        }
    }

    data_min = new double [data_num];
    data_max = new double [data_num];
    
    for ( i = 0; i < data_num; i++ )
    {
        data_min[i] = data_max[i] = data_array[i][0];
        for ( j = 1; j < dataset->dims; j++ )
        {
            if ( data_min[i] > data_array[i][j] )
                data_min[i] = data_array[i][j];
            else if ( data_max[i] < data_array[i][j] )
                data_max[i] = data_array[i][j];
        }
    }
    
}


void CDimReducer::SaveDataToFile(const char * fname)
{
    int i, j;

    FILE *fptr = fopen(fname,"w");
    fprintf(fptr, "%d %d\n", data_num, dataset->dims);

    for ( i = 0; i < data_num; i++ )
    {
        fprintf(fptr, "%lf %lf\n", data_min[i], data_max[i]);
    }

    delete [] data_min;
    delete [] data_max;
    
    for ( j = 0; j < dataset->dims; j++ )
    {
        fprintf(fptr, "%d ", j);
        for ( i = 0; i < data_num; i++ )
        {
            fprintf(fptr, "%lf ", data_array[i][j]);
        }
        fprintf(fptr, "\n");
    }

    for ( i = 0; i < data_num; i++ )
    {
        fprintf(fptr, "%d ", entries_array[i][0]);
    }
    fprintf(fptr, "\n");
    
    fclose(fptr);
}

bool CDimReducer::BuildDimClusterTree()
{
    int N = strlen(dataset->datafile);
	char *dcfile;
    dcfile = new char[N+1];
    strcpy(dcfile, dataset->datafile);
    strcpy(&dcfile[N-4],".dc");
    FILE *fp;
    if ( ( fp = fopen(dcfile, "r" ) ) != NULL )
    {
        fclose(fp);
    }
    else
    {
        if ( !GetData() )
            return false;
        RefineData();
        char* temp_file = "temp__";
        SaveDataToFile(temp_file);
        bool succ = OnlineClusterDimemsion(temp_file, dcfile, 1000, int(dataset->data_size*0.9 ));
#ifdef _WIN32
        DeleteFile(temp_file);
#else
        unlink(temp_file);
#endif
        if ( !succ )
            return false;
    }

    if ( dim_tree != NULL )
        delete dim_tree;
	dim_tree = new CSunburstDimClusterTree();
	dim_tree->BuildSunburstDimClusterTreeFromFile(dcfile);
    if ( representative_dim_list != NULL )
        delete representative_dim_list;
    representative_dim_list = new CLinkList;
    return true;
}

CSunburstDimClusterTree* CDimReducer::GetDimClusterTree()
{
    if ( !BuildDimClusterTree() )
        return NULL;

    return dim_tree;
}

void CDimReducer::AddRepresentativeDim(CSunburstDimCluster* current_node)
{
    if ( dim_tree == NULL )
        return;
    representative_dim_list->AddNodeAtEnd(current_node);    
}


void CDimReducer::DeleteRepresentativeDim(CSunburstDimCluster* current_cluster)
{
    if ( dim_tree == NULL )
        return;
	representative_dim_list->DeleteElement(current_cluster);
}


void CDimReducer::SetStatus(const char *type)
{
    if (strcmp(type,"normal")==0)
    {
        screen_distortion->distortion_status = false; 
        status = NORMAL;
        dataset->CountDims();
        dataset->RemapAllData();
    }
    else if (strcmp(type,"distort")==0)
        screen_distortion->distortion_status = true; 
    else if (strcmp(type,"rotate")==0)
        screen_distortion->distortion_status = true; 
    else if (strcmp(type,"modify")==0)  
    {
        SetStatus("normal");
        status = MODIFY;
        screen_distortion->distortion_status = false; 
    }
    else if (strcmp(type,"select")==0)
    {
//        screen_distortion->distortion_status = false; 
        SetStatus("normal");
        status = SELECT;
    }
    else if (strcmp(type,"customize")==0)
    {
        screen_distortion->distortion_status = false; 
    }
    else if (strcmp(type,"apply")==0)
    {
        screen_distortion->distortion_status = false; 
        if ( status == APPLY )
            return;
        if ( representative_dim_list->GetLength() <= 0 || representative_dim_list->GetLength() > dataset->all_dims )
            return;
        SaveMinMax();
        status = APPLY;
        dataset->dims = representative_dim_list->GetLength();
        dataset->RemapAllData();
    }
}


void CDimReducer::SaveMinMax()
{
    dim_max.SetLen(dataset->all_dims);
    dim_min.SetLen(dataset->all_dims);
    dataset->MapData(dataset->all_dim_max, dim_max);
    dataset->MapData(dataset->all_dim_min, dim_min);
}

void CDimReducer::MapData( ARR<double> all_data, ARR<double> &data)
{
    int i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        data[i] = ((CSunburstDimCluster *)(node->element))->Calculate(all_data);
        node = representative_dim_list->GetNext(node);
        i++;
    }
}


void CDimReducer::MapDataWithDimClusterExtends( ARR<double> all_data, ARR<double> &mean, 
                                                ARR<double> &max, ARR<double> &min)
{
    int i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        ((CSunburstDimCluster *)(node->element))->CalculateWithDimClusterExtends(all_data, mean[i], max[i], min[i]);
        node = representative_dim_list->GetNext(node);
        i++;
    }
}


void CDimReducer::MapData( ARR<int> all_data, ARR<int> &data)
{
    int i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        data[i] = ((CSunburstDimCluster *)(node->element))->Calculate(all_data);
        node = representative_dim_list->GetNext(node);
        i++;
    }
}

void CDimReducer::MapData( ARR<NominalField*> all_nominal_fields, ARR<NominalField*> &nominal_fields)
{
    int i = 0, okc_id;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        okc_id = ((CSunburstDimCluster *)(node->element))->okc_id;
        if ( okc_id >= 0 ) 
            nominal_fields[i] = all_nominal_fields[okc_id];
        else 
            nominal_fields[i] = 0;
        node = representative_dim_list->GetNext(node);
        i++;
    }
}


void CDimReducer::MapData( ARR<char *> all_names, ARR<char *> &names)
{
    int i = 0;
    CLinkListNode *node = representative_dim_list->GetFirst();
    while ( node != NULL )
    {
        names[i] = ((CSunburstDimCluster *)(node->element))->dim_name;
        node = representative_dim_list->GetNext(node);
        i++;
    }
}


void CDimReducer::ExtendNodeInRepresentativeDimList(int N)
{
    if ( status != APPLY )
        return;

    CSunburstDimCluster *cluster;
    CLinkListNode *node;
    cluster = (CSunburstDimCluster *)(representative_dim_list->GetElement(N));
    node = representative_dim_list->Get(N);
    if ( cluster->okc_id >= 0 || cluster->children == NULL )
        return;

    CSunburstDimCluster *child = (CSunburstDimCluster *)cluster->children;
    while ( child != NULL )
    {
        representative_dim_list->DeleteElement(child);
        child->selected = true;
        representative_dim_list->AddNodeAfter(child, node);
        child = (CSunburstDimCluster *)child->siblings;
    }

    representative_dim_list->DeleteNode(node);
    cluster->selected = false;

    dataset->dims = representative_dim_list->GetLength();
    dataset->RemapAllData();
}


void CDimReducer::ShinkNodeInRepresentativeDimList(int N)
{
    if ( status != APPLY )
        return;

    CSunburstDimCluster *cluster;
    CLinkListNode *node;
    cluster = (CSunburstDimCluster *)(representative_dim_list->GetElement(N));
    node = representative_dim_list->Get(N);
    if ( cluster->parent == NULL )
        return;

    CSunburstDimCluster *parent = (CSunburstDimCluster *)cluster->parent;
    parent->selected = true;
    representative_dim_list->AddNodeAfter(parent, node);

    CSunburstDimCluster *child = (CSunburstDimCluster *)parent->children;
    while ( child != NULL )
    {
        representative_dim_list->DeleteElement(child);
        child->selected = false;
        child = (CSunburstDimCluster *)child->siblings;
    }

    dataset->dims = representative_dim_list->GetLength();
    dataset->RemapAllData();
}

double CDimReducer::GetDimDisagreement(int N)
{
    if ( N >= representative_dim_list->GetLength() )
        return 0;

    CSunburstDimCluster *cluster;
    cluster = (CSunburstDimCluster *)(representative_dim_list->GetElement(N));

    return cluster->GetRadius();
}

int CDimReducer::GetEntries(int N)
{
    if ( N >= representative_dim_list->GetLength() )
        return 0;

    CSunburstDimCluster *cluster;
    cluster = (CSunburstDimCluster *)(representative_dim_list->GetElement(N));

    return cluster->entries;
}

bool CDimReducer::IsLeafDim(int N)
{
    if ( N >= representative_dim_list->GetLength() )
        return true;

    CSunburstDimCluster *cluster;
    cluster = (CSunburstDimCluster *)(representative_dim_list->GetElement(N));
    if ( cluster->okc_id < 0 )
        return false;
    else 
        return true;
}


void CDimReducer::SizeStructureSelectNode(int threshold, bool mark)
{
    current_node->ResetSelected( !mark );
    current_node->SizeStructureSelect(threshold, mark);
}


void CDimReducer::RadiusStructureSelectNode(double threshold, bool mark)
{
    current_node->ResetSelected( !mark );
    current_node->RadiusStructureSelect(threshold, mark);
}